import os
from pathlib import Path
import wkw
import numpy as np
import shutil
from os import path, makedirs
import pytest

POSITION = (0, 0, 0)
SIZE = (32, 32, 32)


def test_context_manager():
    wkw.Dataset.create("tests/tmp", wkw.Header(np.uint16)).close()

    with wkw.Dataset.open("tests/tmp") as dataset:
        assert dataset.header.voxel_type == np.uint16


def test_create():
    with wkw.Dataset.create("tests/tmp", wkw.Header(np.uint8)) as dataset:
        assert dataset.header.voxel_type == np.uint8
        assert path.exists(path.join("tests/tmp", "header.wkw"))


def test_open():
    wkw.Dataset.create("tests/tmp", wkw.Header(np.uint16)).close()

    with wkw.Dataset.open("tests/tmp") as dataset:
        assert dataset.header.voxel_type == np.uint16


def test_non_negative_offsets():
    wkw.Dataset.create("tests/tmp", wkw.Header(np.uint8)).close()

    with pytest.raises(AssertionError):
        with wkw.Dataset.open("tests/tmp") as dataset:
            dataset.read((-1, 0, 0), (0, 0, 0))

    with pytest.raises(AssertionError):
        with wkw.Dataset.open("tests/tmp") as dataset:
            dataset.read((0, -1, 0), (0, 0, 0))

    with pytest.raises(AssertionError):
        with wkw.Dataset.open("tests/tmp") as dataset:
            dataset.read((0, 0, -1), (0, 0, 0))

    with pytest.raises(AssertionError):
        with wkw.Dataset.open("tests/tmp") as dataset:
            dataset.read((0, 0, 0), (-1, -1, -1))


def test_empty_read():
    with wkw.Dataset.create("tests/tmp", wkw.Header(np.uint8)) as dataset:
        data = dataset.read((1, 1, 1), (0, 0, 0))
        assert data.shape == (1, 0, 0, 0)


def test_readwrite():
    with wkw.Dataset.create("tests/tmp", wkw.Header(np.uint8)) as dataset:
        header_size = path.getsize(path.join("tests/tmp", "header.wkw"))
        test_data = generate_test_data(dataset.header.voxel_type)

        dataset.write(POSITION, test_data)
        assert (
            path.getsize(path.join("tests/tmp", "z0", "y0", "x0.wkw"))
            == np.prod(SIZE) * (dataset.header.file_len**3) + header_size
        )
        assert np.array_equiv(dataset.read(POSITION, SIZE), test_data)


def test_readwrite_lz4():
    with wkw.Dataset.create(
        "tests/tmp",
        wkw.Header(np.uint8, block_type=wkw.Header.BLOCK_TYPE_LZ4, file_len=4),
    ) as dataset:
        SIZE128 = (128, 128, 128)
        test_data = generate_test_data(dataset.header.voxel_type, SIZE128)

        dataset.write(POSITION, test_data)
        assert np.array_equiv(dataset.read(POSITION, SIZE128), test_data)


def test_readwrite_lz4hc():
    with wkw.Dataset.create(
        "tests/tmp",
        wkw.Header(np.uint8, block_type=wkw.Header.BLOCK_TYPE_LZ4HC, file_len=4),
    ) as dataset:
        SIZE128 = (128, 128, 128)
        test_data = generate_test_data(dataset.header.voxel_type, SIZE128)

        dataset.write(POSITION, test_data)
        assert np.array_equiv(dataset.read(POSITION, SIZE128), test_data)


def test_readwrite_live_compression():
    SIZE128 = (128, 128, 128)
    file_len = 4
    header = wkw.Header(
        np.uint8, block_type=wkw.Header.BLOCK_TYPE_LZ4, file_len=file_len
    )
    with wkw.Dataset.create("tests/tmp", header) as dataset:
        header_size = path.getsize(path.join("tests/tmp", "header.wkw"))
        test_data = generate_test_data(dataset.header.voxel_type, SIZE128)

        dataset.write(POSITION, test_data)

        # The size should be less than if it was not compressed
        assert (
            path.getsize(path.join("tests/tmp", "z0", "y0", "x0.wkw"))
            < np.prod(SIZE128) * (dataset.header.file_len**3) + header_size
        )

    with wkw.Dataset.open("tests/tmp") as dataset:
        assert np.array_equiv(dataset.read(POSITION, SIZE128), test_data)


def test_readwrite_live_compression_should_enforce_full_file_write():
    with pytest.raises(Exception):
        with wkw.Dataset.create(
            "tests/tmp", wkw.Header(np.uint8, block_type=wkw.Header.BLOCK_TYPE_LZ4)
        ) as dataset:
            test_data = generate_test_data(dataset.header.voxel_type)
            dataset.write(POSITION, test_data)


def test_readwrite_live_compression_should_not_allow_inconsistent_writes():
    SIZE129 = (129, 128, 128)
    file_len = 4
    header = wkw.Header(
        np.uint8, block_type=wkw.Header.BLOCK_TYPE_LZ4, file_len=file_len
    )
    test_data = generate_test_data(header.voxel_type, SIZE129)
    empty_data = np.zeros(SIZE129).astype(header.voxel_type)

    with wkw.Dataset.create("tests/tmp", header) as dataset:
        with pytest.raises(Exception):
            dataset.write(POSITION, test_data)

    with wkw.Dataset.open("tests/tmp") as dataset:
        assert np.array_equiv(dataset.read(POSITION, SIZE129), empty_data)


def test_readwrite_live_compression_should_truncate():
    SIZE128 = (128, 128, 128)
    file_len = 4
    header = wkw.Header(
        np.uint8, block_type=wkw.Header.BLOCK_TYPE_LZ4, file_len=file_len
    )
    test_data = generate_test_data(header.voxel_type, SIZE128)
    ones_data = np.ones(SIZE128).astype(header.voxel_type)

    with wkw.Dataset.create("tests/tmp", header) as dataset:
        dataset.write(POSITION, test_data)

    random_compressed_size = path.getsize(path.join("tests/tmp", "z0", "y0", "x0.wkw"))

    with wkw.Dataset.open("tests/tmp") as dataset:
        dataset.write(POSITION, ones_data)

    empty_compressed_size = path.getsize(path.join("tests/tmp", "z0", "y0", "x0.wkw"))

    assert empty_compressed_size < random_compressed_size

    with wkw.Dataset.open("tests/tmp") as dataset:
        assert np.array_equiv(dataset.read(POSITION, SIZE128), ones_data)


def test_compress():
    with wkw.Dataset.create("tests/tmp", wkw.Header(np.uint8)) as dataset:
        test_data = generate_test_data(dataset.header.voxel_type)
        dataset.write(POSITION, test_data)

        with dataset.compress("tests/tmp2", compress_files=True) as dataset2:
            assert dataset2.header.voxel_type == np.uint8
            assert dataset2.header.block_type == wkw.Header.BLOCK_TYPE_LZ4HC

            header_size = path.getsize(path.join("tests/tmp2", "header.wkw"))

            assert path.exists(path.join("tests/tmp2", "header.wkw"))
            assert (
                path.getsize(path.join("tests/tmp2", "z0", "y0", "x0.wkw"))
                < np.prod(SIZE) * (dataset2.header.file_len**3) + header_size
            )
            assert np.array_equiv(dataset2.read(POSITION, SIZE), test_data)


def test_row_major_order():
    data_shape = (4, 5, 6)
    data = generate_test_data(np.uint8, data_shape)
    with wkw.Dataset.create("tests/tmp", wkw.Header(np.uint8)) as dataset:
        dataset.write((0, 0, 0), data)
        read_data = dataset.read((0, 0, 0), data_shape)

    assert np.array_equiv(data, read_data)

    with wkw.Dataset.create("tests/tmp2", wkw.Header(np.uint8)) as dataset:
        fortran_data = np.asfortranarray(data)
        dataset.write((0, 0, 0), fortran_data)
        fortran_read_data = dataset.read((0, 0, 0), data_shape)

    assert np.array_equiv(fortran_read_data, read_data)
    assert np.array_equiv(fortran_read_data, fortran_data)


def test_row_major_order_with_offset():
    data_shape = (17, 1, 4)
    data = generate_test_data(np.uint8, data_shape)
    with wkw.Dataset.create("tests/tmp", wkw.Header(np.uint8)) as dataset:
        dataset.write((15, 2, 0), data)
        read_data = dataset.read((15, 2, 0), data_shape)

    assert np.array_equiv(data, read_data)


def test_row_major_order_with_different_voxel_size():
    data_shape = (4, 3, 9)
    data = generate_test_data(np.uint16, data_shape)
    with wkw.Dataset.create("tests/tmp", wkw.Header(np.uint16)) as dataset:
        dataset.write((3, 1, 0), data)
        read_data = dataset.read((3, 1, 0), data_shape)

    assert np.array_equiv(data, read_data)


def test_row_major_order_with_channels():
    data_shape = (2, 4, 3, 9)
    data = generate_test_data(np.uint8, data_shape)
    with wkw.Dataset.create(
        "tests/tmp", wkw.Header(np.uint8, num_channels=2)
    ) as dataset:
        dataset.write((3, 1, 0), data)
        read_data = dataset.read((3, 1, 0), data_shape[1:])

    assert np.array_equiv(data, read_data)


def test_row_major_order_with_channels_and_different_voxel_size():
    data_shape = (2, 4, 3, 9)
    data = generate_test_data(np.uint16, data_shape)
    with wkw.Dataset.create(
        "tests/tmp", wkw.Header(np.uint16, num_channels=2)
    ) as dataset:
        dataset.write((3, 1, 0), data)
        read_data = dataset.read((3, 1, 0), data_shape[1:])

    assert np.array_equiv(data, read_data)


def test_column_major_order_with_channels_and_different_voxel_size():
    data_shape = (2, 4, 3, 9)
    data = generate_test_data(np.uint16, data_shape, order="F")
    with wkw.Dataset.create(
        "tests/tmp", wkw.Header(np.uint16, num_channels=2)
    ) as dataset:
        dataset.write((3, 1, 0), data)
        read_data = dataset.read((3, 1, 0), data_shape[1:])

    assert np.array_equiv(data, read_data)


def test_view_on_np_array():
    data_shape = (4, 4, 9)
    data = generate_test_data(np.uint16, data_shape)
    data = data[:, ::2]
    with wkw.Dataset.create("tests/tmp", wkw.Header(np.uint16)) as dataset:
        dataset.write((3, 1, 0), data)
        read_data = dataset.read((3, 1, 0), data.shape)

    assert np.array_equiv(data, read_data)


def test_not_too_much_data_is_written():
    def write_and_test_in_given_order(wkw_path, order):
        data_shape = (35, 35, 35)
        data = generate_test_data(np.uint8, data_shape, order=order)
        with wkw.Dataset.create(wkw_path, wkw.Header(np.uint8)) as dataset:
            dataset.write((0, 0, 0), np.ones((35, 35, 64), dtype=np.uint8))
            dataset.write((1, 2, 3), data)

            read_data = dataset.read((1, 2, 3), (35, 35, 35))
            before = dataset.read((0, 0, 0), (1, 2, 3))
            after = dataset.read((0, 0, 38), (35, 35, 26))

        assert np.array_equiv(data, read_data)
        assert np.array_equiv(before, 1)
        assert np.array_equiv(after, 1)

    write_and_test_in_given_order("tests/tmp", "F")
    write_and_test_in_given_order("tests/tmp2", "C")


def test_multiple_writes_and_reads():
    mem_buffer = np.zeros((200, 200, 200), dtype=np.uint8, order="F")
    with wkw.Dataset.create("tests/tmp", wkw.Header(np.uint8)) as dataset:
        for i in range(10):
            offset = np.random.randint(100, size=(3))
            size = np.random.randint(1, 100, size=(3))
            order = np.random.choice(["F", "C"])
            data = generate_test_data(np.uint8, [1] + list(size), order)
            dataset.write(offset, data)
            mem_buffer[
                offset[0] : offset[0] + size[0],
                offset[1] : offset[1] + size[1],
                offset[2] : offset[2] + size[2],
            ] = data

            read_data = dataset.read((0, 0, 0), (200, 200, 200))
            assert np.array_equiv(mem_buffer, read_data)


def test_multi_channel_column_major_order():
    with wkw.Dataset.create(
        "tests/tmp", wkw.Header(np.uint8, num_channels=3)
    ) as dataset:
        offset = (30, 20, 10)
        data_shape = (3, 100, 200, 300)
        order = "C"
        data = generate_test_data(np.uint8, list(data_shape), order)
        dataset.write(offset, data)

        read_data = dataset.read(offset, data_shape[1:])
        assert np.array_equiv(data, read_data)


def test_big_read():
    data = np.ones((10, 10, 764), order="C", dtype=np.uint8)
    offset = np.array([0, 0, 640])
    bottom = (1200, 2000, 2000)

    with wkw.Dataset.create("tests/tmp", wkw.Header(np.uint8)) as dataset:
        dataset.write(offset, data)
        read_data = dataset.read((0, 0, 0), bottom)[0]
        assert np.all(
            read_data[
                offset[0] : (offset[0] + data.shape[0]),
                offset[1] : (offset[1] + data.shape[1]),
                offset[2] : (offset[2] + data.shape[2]),
            ]
            == 1
        )
        assert np.count_nonzero(read_data[: offset[0], :, :]) == 0
        assert np.count_nonzero(read_data[offset[0] + data.shape[0] :, :, :]) == 0
        assert np.count_nonzero(read_data[:, : offset[1], :]) == 0
        assert np.count_nonzero(read_data[:, offset[1] + data.shape[1] :, :]) == 0
        assert np.count_nonzero(read_data[:, :, : offset[2]]) == 0
        assert np.count_nonzero(read_data[:, :, offset[2] + data.shape[2] :]) == 0


def test_invalid_dataset():
    with pytest.raises(wkw.wkw.WKWException) as excinfo:
        with wkw.Dataset.open("/path/does/not/exist") as dataset:
            pass
    print(excinfo.value)

    with pytest.raises(wkw.wkw.WKWException) as excinfo:
        makedirs("tests/tmp/exists", exist_ok=True)
        with wkw.Dataset.open("tests/tmp/exists") as dataset:
            pass
    print(excinfo.value)


def test_dataset_with_invalid_jumptable():
    with wkw.Dataset.create(
        "tests/tmp",
        wkw.Header(np.uint8, block_type=wkw.Header.BLOCK_TYPE_LZ4HC, file_len=4),
    ) as dataset:
        test_data = generate_test_data(
            dataset.header.voxel_type, (32 * 4, 32 * 4, 32 * 4)
        )
        dataset.write((0, 0, 0), test_data)

    # reset jumptable with zeros
    data_path = Path("tests/tmp") / "z0" / "y0" / "x0.wkw"
    wkw_file_bytes = bytearray(data_path.read_bytes())
    zeros = b"\x00" * (4**3 * 8)
    wkw_file_bytes[16 : (16 + (4**3) * 8)] = zeros
    data_path.write_bytes(wkw_file_bytes)

    with pytest.raises(wkw.wkw.WKWException) as excinfo:
        with wkw.Dataset.open("tests/tmp") as dataset:
            dataset.read((0, 0, 0), (32 * 4,) * 3)

    print(excinfo)
    assert "Corrupt jump table" in str(excinfo)


def test_readwrite_compressed_tmp():
    with wkw.Dataset.create(
        "tests/tmp",
        wkw.Header(np.uint8, block_type=wkw.Header.BLOCK_TYPE_LZ4, file_len=1),
    ) as dataset:
        test_data = generate_test_data(dataset.header.voxel_type)

        wkw_path = path.join("tests/tmp", "z0", "y0", "x0.wkw")

        # write data, should run through cleanly
        dataset.write(POSITION, test_data)
        assert np.array_equiv(dataset.read(POSITION, SIZE), test_data)
        assert path.exists(wkw_path)
        assert not path.exists(wkw_path + "_tmp")

        # simulate broken write
        os.rename(wkw_path, wkw_path + "_tmp")
        wkw_file_bytes = bytearray(Path(wkw_path + "_tmp").read_bytes())
        zeros = b"\x00" * 8
        wkw_file_bytes[16 : (16 + 8)] = zeros
        Path(wkw_path + "_tmp").write_bytes(wkw_file_bytes)

        # try another write, should clean up broken write
        dataset.write(POSITION, test_data)
        assert path.exists(wkw_path)
        assert not path.exists(wkw_path + "_tmp")
        assert np.array_equiv(dataset.read(POSITION, SIZE), test_data)


def generate_test_data(dtype, size=SIZE, order="C"):
    return np.array(
        np.random.uniform(np.iinfo(dtype).min, np.iinfo(dtype).max, size).astype(dtype),
        order=order,
    )


def try_rmtree(dir):
    try:
        shutil.rmtree(dir)
    except FileNotFoundError:
        pass


def setup_function():
    np.random.seed(0)


def teardown_function():
    try_rmtree("tests/tmp")
    try_rmtree("tests/tmp2")



import cffi
import ctypes
import numpy as np
import platform
import os
from copy import deepcopy
from glob import iglob


def _init_libwkw():
    this_dir = os.path.dirname(__file__)
    path_wkw_header = os.path.join(this_dir, "lib", "wkw.h")

    lib_name_platform = {
        "Linux": "libwkw.so",
        "Windows": "wkw.dll",
        "Darwin": "libwkw.dylib",
    }
    path_wkw_lib = os.path.join(this_dir, "lib", lib_name_platform[platform.system()])

    with open(path_wkw_header) as f:
        wkw_header = f.readlines()

        # strip away directives to be compatible with cffi module
        wkw_header = filter(lambda l: not l.startswith("#"), wkw_header)
        wkw_header = "\n".join(wkw_header)

    ffi = cffi.FFI()
    ffi.cdef(wkw_header)
    libwkw = ffi.dlopen(path_wkw_lib)

    return (ffi, libwkw)


ffi, libwkw = _init_libwkw()


class WKWException(Exception):
    pass


def _raise_wkw_exception():
    error_msg = ffi.string(libwkw.get_last_error_msg())
    raise WKWException(error_msg.decode())


def _check_wkw(ret):
    if ret > 0:
        _raise_wkw_exception()


def _check_wkw_null(ret):
    if ret == ffi.NULL:
        _raise_wkw_exception()
    return ret


class Header:
    BLOCK_TYPE_RAW = 1
    BLOCK_TYPE_LZ4 = 2
    BLOCK_TYPE_LZ4HC = 3

    VALID_BLOCK_TYPES = [BLOCK_TYPE_RAW, BLOCK_TYPE_LZ4, BLOCK_TYPE_LZ4HC]
    VALID_VOXEL_TYPES = [
        np.uint8,
        np.uint16,
        np.uint32,
        np.uint64,
        np.float32,
        np.float64,
        np.int8,
        np.int16,
        np.int32,
        np.int64,
    ]

    def __init__(
        self,
        voxel_type: type,
        num_channels: int = 1,
        version: int = 1,
        block_len: int = 32,
        file_len: int = 32,
        block_type: int = 1,
    ):
        self.version = version

        assert (
            block_len & (block_len - 1) == 0
        ), f"block_len must be a power of two, but is {block_len}"
        self.block_len = block_len

        assert (
            file_len & (file_len - 1) == 0
        ), f"file_len must be a power of two, but is {file_len}"
        self.file_len = file_len

        assert (
            block_type in self.VALID_BLOCK_TYPES
        ), f"block_type must be one of {self.VALID_BLOCK_TYPES}, but is {block_type}"
        self.block_type = block_type

        assert (
            voxel_type in self.VALID_VOXEL_TYPES
        ), f"voxel_type must be one of {self.VALID_VOXEL_TYPES}, but is {voxel_type}"
        self.voxel_type = voxel_type

        assert num_channels > 0, f"num_channels must be >0, but is {num_channels}"
        self.num_channels = num_channels

    @staticmethod
    def from_c(header_c):
        assert header_c.voxel_type > 0
        voxel_type = Header.VALID_VOXEL_TYPES[header_c.voxel_type - 1]
        voxel_type_size = np.dtype(voxel_type).itemsize

        assert header_c.voxel_size % voxel_type_size == 0
        num_channels = header_c.voxel_size // voxel_type_size

        assert header_c.block_type > 0
        block_type = Header.VALID_BLOCK_TYPES[header_c.block_type - 1]

        return Header(
            version=header_c.version,
            block_len=header_c.block_len,
            file_len=header_c.file_len,
            block_type=block_type,
            voxel_type=voxel_type,
            num_channels=num_channels,
        )

    def to_c(self):
        voxel_type_c = Header.VALID_VOXEL_TYPES.index(self.voxel_type) + 1
        voxel_type_size = np.dtype(self.voxel_type).itemsize
        voxel_size = self.num_channels * voxel_type_size

        header_c = ffi.new("struct header *")
        header_c.version = self.version
        header_c.file_len = self.file_len
        header_c.block_len = self.block_len
        header_c.block_type = self.block_type
        header_c.voxel_type = voxel_type_c
        header_c.voxel_size = voxel_size
        return header_c


def _build_box(off, shape):
    assert off[0] >= 0, "Offset x must be greater or equal to 0"
    assert off[1] >= 0, "Offset y must be greater or equal to 0"
    assert off[2] >= 0, "Offset z must be greater or equal to 0"
    assert shape[0] >= 0, "Shape x must be greater or equal to 0"
    assert shape[1] >= 0, "Shape y must be greater or equal to 0"
    assert shape[2] >= 0, "Shape z must be greater or equal to 0"

    off = np.asarray(off, dtype=np.uint32)
    shape = np.asarray(shape, dtype=np.uint32)
    return np.hstack((off, off + shape))


class File:
    @staticmethod
    def compress(src_path: str, dst_path: str):
        src_path_c = ffi.new("char[]", src_path.encode())
        dst_path_c = ffi.new("char[]", dst_path.encode())

        _check_wkw(libwkw.file_compress(src_path_c, dst_path_c))


class Dataset:
    def __init__(self, root, handle):
        self.root = ffi.string(root).decode("utf-8")
        self.handle = handle

        header_c = ffi.new("struct header *")
        libwkw.dataset_get_header(self.handle, header_c)
        self.header = Header.from_c(header_c)

    def read(self, off, shape):
        box = _build_box(off, shape)
        box_ptr = ffi.cast("uint32_t *", box.ctypes.data)

        num_channels = self.header.num_channels
        data = np.zeros(
            (num_channels,) + tuple(shape), order="F", dtype=self.header.voxel_type
        )
        data_ptr = ffi.cast("void *", data.ctypes.data)

        _check_wkw(libwkw.dataset_read(self.handle, box_ptr, data_ptr))
        return data

    def write(self, off, data):
        if not isinstance(data, np.ndarray):
            raise WKWException("Data must be a NumPy ndarray")

        if not data.ndim in [3, 4]:
            raise WKWException("Data must be three- or four-dimensional")

        data = data.reshape((-1,) + data.shape[-3:])
        if not data.shape[0] == self.header.num_channels:
            raise WKWException(
                "Data volume must have {} channels".format(self.header.num_channels)
            )

        if not data.dtype == self.header.voxel_type:
            raise WKWException(
                "Data elements must be of type {}".format(self.header.voxel_type)
            )

        def is_contiguous(data):
            return data.flags["F_CONTIGUOUS"] or data.flags["C_CONTIGUOUS"]

        if not is_contiguous(data):
            data = np.asfortranarray(data)

        box = _build_box(off, data.shape[-3:])
        box_ptr = ffi.cast("uint32_t *", box.ctypes.data)

        assert is_contiguous(data), "Input data is not contiguous"

        data_in_c_order = data.flags["C_CONTIGUOUS"]
        data_ptr = ffi.cast("void *", data.ctypes.data)
        _check_wkw(
            libwkw.dataset_write(self.handle, box_ptr, data_ptr, data_in_c_order)
        )

    def compress(self, dst_path: str, compress_files: bool = False):
        header = deepcopy(self.header)
        header.block_type = Header.BLOCK_TYPE_LZ4HC

        dataset = Dataset.create(dst_path, header)

        if compress_files:
            for file in self.list_files():
                rel_file = os.path.relpath(file, self.root)
                File.compress(file, os.path.join(dst_path, rel_file))

        return dataset

    def list_files(self):
        return iglob(os.path.join(self.root, "*", "**", "*.wkw"))

    def close(self):
        libwkw.dataset_close(self.handle)

    @staticmethod
    def open(root: str, header=None):
        root_c = ffi.new("char[]", root.encode())
        handle = None
        if header is not None:
            try:
                handle = _check_wkw_null(libwkw.dataset_create(root_c, header.to_c()))
            except WKWException:
                handle = _check_wkw_null(libwkw.dataset_open(root_c))
        else:
            handle = _check_wkw_null(libwkw.dataset_open(root_c))
        return Dataset(root_c, handle)

    @staticmethod
    def create(root: str, header):
        root_c = ffi.new("char[]", root.encode())
        handle = _check_wkw_null(libwkw.dataset_create(root_c, header.to_c()))
        return Dataset(root_c, handle)

    def __enter__(self):
        return self

    def __exit__(self, type, value, tb):
        self.close()



from setuptools import setup, find_packages
from setuptools.command.build_py import build_py

import os
import subprocess
import shutil
import platform


class BuildPyCommand(build_py):
    """Modified build command to compile C library."""

    def __build_c_library(self):
        this_dir = os.path.dirname(__file__)
        c_dir = os.path.normpath(os.path.join(this_dir, "..", "c"))

        # building C library
        subprocess.call(["cargo", "clean"], cwd=c_dir)
        if platform.system() == "Darwin":
            print("x86_64-apple-darwin compilation")
            subprocess.call(
                ["cargo", "build", "--target=x86_64-apple-darwin", "--release"],
                cwd=c_dir,
            )
            print("aarch64-apple-darwin compilation")
            subprocess.call(
                ["cargo", "build", "--target=aarch64-apple-darwin", "--release"],
                cwd=c_dir,
            )
            subprocess.call(
                [
                    "lipo",
                    "-create",
                    os.path.join(
                        c_dir,
                        "target",
                        "aarch64-apple-darwin",
                        "release",
                        "libwkw.dylib",
                    ),
                    os.path.join(
                        c_dir,
                        "target",
                        "x86_64-apple-darwin",
                        "release",
                        "libwkw.dylib",
                    ),
                    "-output",
                    os.path.join(c_dir, "target", "release", "libwkw.dylib"),
                ],
                cwd=c_dir,
            )
        else:
            subprocess.call(["cargo", "build", "--release"], cwd=c_dir)

        lib_name_platform = {
            "Linux": "libwkw.so",
            "Windows": "wkw.dll",
            "Darwin": "libwkw.dylib",
        }
        lib_name = lib_name_platform[platform.system()]
        lib_file = os.path.join(c_dir, "target", "release", lib_name)
        header_file = os.path.join(c_dir, "include", "wkw.h")

        # copying to lib dir
        lib_dir = os.path.join(this_dir, "wkw", "lib")

        if os.path.exists(lib_dir):
            shutil.rmtree(lib_dir)

        os.makedirs(lib_dir)
        shutil.copy(lib_file, os.path.join(lib_dir, lib_name))
        shutil.copy(header_file, os.path.join(lib_dir, "wkw.h"))

    def run(self):
        self.__build_c_library()
        build_py.run(self)


setup(
    name="wkw",
    use_scm_version={"root": ".."},
    setup_requires=["setuptools_scm"],
    author="Alessandro Motta",
    author_email="alessandro.motta@brain.mpg.de",
    url="https://github.com/scalableminds/webknossos-wrap",
    long_description_content_type="text/markdown",
    packages=find_packages(),
    include_package_data=True,
    license="MIT",
    install_requires=["cffi", "numpy>=1.15,==1.*"],
    extras_require={
        ':python_version < "3.8"': [
            "numpy<1.22",
        ],
    },
    cmdclass={"build_py": BuildPyCommand},
)
